import java.util.*;
import javafx.scene.SubScene;
import javafx.scene.Group;
import javafx.scene.layout.StackPane;
import javafx.scene.Node;
import javafx.collections.ObservableList;
import javafx.event.EventHandler;
import javafx.event.Event;
import javafx.event.ActionEvent;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.KeyCode;
import javafx.animation.Timeline;
import javafx.animation.KeyFrame;
import javafx.util.Duration;

/**
 * パネルクラス
 * <pre>
 * 時間の制御を行い、適宜、スプライトグループ、スプライトを呼び出します。
 * キーの押下状態を制御します。
 * StackPaneを利用することで、描画順を保証します。
 * </pre>
 */
public class KPanel extends StackPane implements KConstant {

	/** 制御クラス */
	private KCtrl ctrl = null;

	/** 比率 */
	private double ratio = 1f;

	/** 画面伸縮モード */
	private boolean stretchScreen = false;

	/** スプライトグループ一覧 */
	public ArrayList<KSpriteGroup> spriteGroupLst = null;

	/** スプライト層深さ */
	private int spriteLayoutCount = 0;

	/** キーを押している状態一覧 */
	private HashMap<KeyCode, String> mapKeyPress = null;

	/** 時間 */
	private int time = 0;

	// ゲーム中フラグ
	boolean flgPlaying = true;

	/**
	 * コンストラクタ
	 * @param ctrl 制御クラス
	 */
	public KPanel(KCtrl ctrl){

		// スーパークラスを呼び出す
		super();

		// 制御クラス保持
		this.ctrl = ctrl;

		// 初期化処理を呼び出す
		init();

	} // end KPanel

	/**
	 * 初期化処理
	 * <pre>
	 * パネルクラスの初期化処理を行います。
	 * </pre>
	 */
	public void init(){

		// キー押下状況一覧を生成
		mapKeyPress = new HashMap<KeyCode, String>();

		// キーを押したイベントを登録
		this.addEventHandler(KeyEvent.KEY_PRESSED, event -> KPanelKeyPressed(event));

		// キーを離したイベントを登録
		this.addEventHandler(KeyEvent.KEY_RELEASED, event -> KPanelKeyKeyReleased(event));

		// スプライトグループ一覧生成
		spriteGroupLst = new ArrayList<KSpriteGroup>(100);

	} // end init

	/**
	 * スプライト層設定
	 * <pre>
	 * スプライト層の深さの設定をします。
	 * 描画順が保証されます（SKKFXはJavaFXの機能を利用）。
	 * </pre>
	 * @param layoutCount スプライト層の深さ
	 */
	public void setLayoutCount(int layoutCount){

		// スプライト層深さを保持
		spriteLayoutCount = layoutCount;

		// スプライト層分、ループ
		for(int i = 0; i < layoutCount; i++){

			// グループを生成
			Group group = new Group();

			// スプライト一覧を生成
			SubScene spriteLst = new SubScene(group, WIDTH, HEIGHT);

			// スプライト層一覧に追加
			this.getChildren().add(spriteLst);

		} // end for スプライト層分、ループ

	} // end setLayoutCount

	/**
	 * スプライト層初期化
	 * <pre>
	 * スプライト層に登録されているスプライトを削除します。
	 * 元々登録されていた場合には、クリアされます。
	 * </pre>
	 * @param layout 初期化するスプライト層
	 */
	public void initKSpriteLstInLayout(int layout){

		// グループを生成
		Group group = new Group();

		// スプライト一覧を生成
		SubScene spriteLst = new SubScene(group, WIDTH, HEIGHT);

		// スプライト層一覧に設定
		this.getChildren().set(layout, spriteLst);

	} // end initKSpriteLstInLayout

	/**
	 * スプライトグループ初期化
	 * <pre>
	 * スプライトグループを初期化します。
	 * 元々登録されていた場合には、クリアされます。
	 * </pre>
	 */
	public void initSpriteGroup(){

		// スプライトグループ一覧生成
		this.spriteGroupLst = new ArrayList<KSpriteGroup>(100);

	}

	/**
	 * 開始
	 * <pre>
	 * 当オブジェクトの処理を開始します。
	 * タイマーが動き始め、コントロールクラス および 各スプライトオブジェクトの処理が定期的に呼び出されます。
	 * </pre>
	 */
	public void start(){

		// タイマーをスタート
		Timeline timer = new Timeline(new KeyFrame(Duration.millis(4), event -> run()));
		timer.setCycleCount(Timeline.INDEFINITE);
		timer.play();

	}

	/**
	 * 実行
	 * <pre>
	 * タイマーから定期的に呼ばれます。
	 * </pre>
	 */
	synchronized public void run(){

		// 時間＋１
		time++;

		// 制御の時間
		if(time % 4 == 0){

			// 制御クラスの実行メソッド呼び出し
			flgPlaying = ctrl.run();

			// ゲーム中の場合（ポーズ中ではない場合）、
			if(flgPlaying){

				//
				// スプライトグループ処理
				//

				// スプライトグループ一覧の件数を取得
				int countSpriteGroupLst = spriteGroupLst.size();

				// スプライトグループ一覧の件数分、ループ
				for(int i = 0; i < countSpriteGroupLst; i++){

					// スプライトグループを取得
					KSpriteGroup spriteGroup = spriteGroupLst.get(i);

					// スプライトグループがnull以外の場合、
					if(spriteGroup != null){

						// スプライトグループの時間を進める
						spriteGroup.forward();

						// スプライトグループを実行
						spriteGroup.run();

					}

				} // end for スプライトグループ一覧の件数分、ループ

				// スプライトグループ一覧の件数が100件を超えた場合、
				if(countSpriteGroupLst > 100){

					// スプライトグループ一覧の件数分、ループ
					for(int i = 0; i < countSpriteGroupLst; i++){

						// スプライトグループを取得
						KSpriteGroup spriteGroup = spriteGroupLst.get(i);

						// スプライトグループがnullの場合、
						if(spriteGroup == null){

							// スプライトグループをつめる
							spriteGroupLst.remove(i);

							// インデックス戻し
							i--;
							countSpriteGroupLst--;

						}

					} // end for スプライトグループ一覧の件数分、ループ

				} // end if スプライトグループ一覧の件数が100件を超えた場合

				//
				// スプライト処理
				//

				// スプライト層深さ分、ループ
				for(int i = 0; i < spriteLayoutCount; i++){

					// スプライト一覧を取得
					List<Node> spriteLst = this.getSpriteLst(i);

					// スプライト一覧の件数を取得
					int spriteLstCount = spriteLst.size();

					// スプライト一覧の件数分、ループ
					for(int j = 0; j < spriteLstCount; j++){

						// オブジェクトを取得
						Node node = spriteLst.get(j);

						// ダミースプライトの場合、
						if(node instanceof KSpriteDummy){

							// 何もしない

						// スプライトの場合、
						}else if(node instanceof KSprite){

							// スプライトを取得
							KSprite sprite = (KSprite)node;

							// スプライトを起動
							sprite.run();

						// テキストの場合、
						}else if(node instanceof KText){

							// スプライトを取得
							KText text = (KText)node;

							// テキストを起動
							text.run();

						// キャンバスの場合、
						}else if(node instanceof KCanvas){

							// キャンバスを取得
							KCanvas canvas = (KCanvas)node;

							// テキストを起動
							canvas.run();

						}

					} // end for スプライト一覧の件数分、ループ

				} // end for スプライト層深さ分、ループ

			} // end if ゲーム中の場合（ポーズ中ではない場合）

		// お掃除の時間
		}else if(time % 4 == 2){

			// ゲーム中の場合（ポーズ中ではない場合）、
			if(flgPlaying){

				//
				// スプライト一覧が多くなった場合は、お掃除をする。（メモリ増加防止）
				// スプライトの実行処理とは別にする（処理分散）
				//

				// スプライト層深さ分、ループ
				for(int i = 0; i < spriteLayoutCount; i++){

					// スプライト一覧を取得
					ObservableList<Node> spriteLst = this.getSpriteLst(i);

					// スプライト一覧の件数を取得
					int spriteLstCount = spriteLst.size();

					// スプライト一覧の件数が600件を超えた場合、
					if(spriteLstCount > 600){

						// 一時的にスプライト一覧（ワーク）を作成
						ArrayList<Node> spriteLstWk = new ArrayList<Node>(1000);

						// スプライト一覧の件数分、ループ
						for(int j = 0; j < spriteLstCount; j++){

							// ノードを取得
							Node node = (Node)spriteLst.get(j);

							// ノードがダミースプライト以外の場合、
							if(!(node instanceof KSpriteDummy)){

								// スプライト一覧（ワーク）に追加
								spriteLstWk.add(node);

							}

						} // end for スプライト一覧の件数分、ループ

						// スプライト一覧を入れ替える
						spriteLst.setAll(spriteLstWk);

					} // end if スプライト一覧の件数が600件を超えた場合

				} // end for スプライト層深さ分、ループ

			} // end if ゲーム中の場合（ポーズ中ではない場合）

		} // end if お掃除の時間

	} // end run

	/**
	 * スプライトグループ追加
	 * @param spriteGroup スプライト
	 */
	public void addSpriteGroup(KSpriteGroup spriteGroup){

		// スプライトグループを追加
		spriteGroupLst.add(spriteGroup);

	} // end addSpriteGroup

	/**
	 * スプライトグループ削除
	 * @param spriteGroup スプライト
	 */
	public void removeSpriteGroup(KSpriteGroup spriteGroup){

		// インデックスを求める
		int index = spriteGroupLst.indexOf(spriteGroup);

		// インデックスが取得できた場合、
		if(index != -1){

			// スプライトグループを削除
			spriteGroupLst.set(index, null);

		}

	} // end addSpriteGroup

	/**
	 * スプライト追加
	 * @param layout スプライト層
	 * @param sprite スプライト
	 */
	public void addSprite(int layout, KSprite sprite){

		// スプライト一覧を取得
		List<Node> spriteLst = this.getSpriteLst(layout);

		// スプライトを追加
		spriteLst.add(sprite);

	} // end addSprite

	/**
	 * スプライト削除
	 * <pre>
	 * スプライト層から探索して、スプライトを削除します。
	 * 当メソッドでは、ダミースプライトに置き換え、runメソッドの中で削除します。
	 * nullに置き換えができない
     * かつ Listクラスにremoveメソッドを多用することで配列コピーを多用されることを避けるため
	 * （重くなることをできるだけ避ける）
	 * </pre>
	 * @param layout スプライト層
	 * @param sprite スプライト
	 */
	public void removeSprite(int layout, KSprite sprite){

		// スプライト一覧を取得
		List<Node> spriteLst = this.getSpriteLst(layout);

		// インデックスを求める
		int index = spriteLst.indexOf(sprite);

		// インデックスが取得できた場合、
		if(index != -1){

			// ダミースプライトに置き換え
			spriteLst.set(index, new KSpriteDummy(this));

		}

	} // end removeSprite

	/**
	 * スプライト削除
	 * <pre>
	 * 全てのスプライト層を探索して、スプライトを削除します。
	 * 通常はスプライト層が引数として設定しているメソッドを呼んでください。
	 * </pre>
	 * @param sprite スプライト
	 */
	public void removeSprite(KSprite sprite){

		// スプライト層深さ分、ループ
		for(int i = 0; i < spriteLayoutCount; i++){

			// スプライト削除
			removeSprite(i, sprite);

		}

	} // end addSprite

	/**
	 * スプライト一覧取得
	 * @param layout スプライト層
	 * @return ArrayList スプライト一覧
	 */
	public ObservableList<Node> getSpriteLst(int layout){

		// スプライト一覧を取得
		SubScene subscene = (SubScene)this.getChildren().get(layout);

		// Groupを取得
		Group group = (Group)subscene.getRoot();

		// Listを取得
		ObservableList<Node> spriteLst = group.getChildren();

		// リターン
		return spriteLst;

	} // end getSpriteLst

	/**
	 * テキスト追加
	 * @param layout スプライト層
	 * @param kText テキストクラス
	 */
	public void addText(int layout, KText kText){

		// スプライト一覧を取得
		List<Node> spriteLst = this.getSpriteLst(layout);

		// テキストクラスを追加
		spriteLst.add(kText);

	} // end addText

	/**
	 * テキスト削除
	 * <pre>
	 * スプライト層から探索して、テキストを削除します。
	 * 当メソッドでは、ダミースプライトに置き換え、runメソッドの中で削除します。
	 * </pre>
	 * @param layout スプライト層
	 * @param kText テキストクラス
	 */
	public void removeText(int layout, KText kText){

		// スプライト一覧を取得
		List<Node> spriteLst = this.getSpriteLst(layout);

		// インデックスを取得
		int index = spriteLst.indexOf(kText);

		// インデックスが取得できた場合、
		if(index != -1){

			// ダミースプライトに置き換え
			spriteLst.set(index, new KSpriteDummy(this));

		}

	} // end removeText

	/**
	 * メディアビュー追加
	 * @param layout スプライト層
	 * @param sprite スプライト
	 */
	public void addMediaView(int layout, KSoundMP3 mp3){

		// スプライト一覧を取得
		List<Node> spriteLst = this.getSpriteLst(layout);

		// メディアビューを追加
		spriteLst.add(mp3);

	} // end addMediaView

	/**
	 * メディアビュー削除
	 * <pre>
	 * スプライト層から探索して、スプライトを削除します。
	 * 当メソッドでは、ダミースプライトに置き換え、runメソッドの中で削除します。
	 * nullに置き換えができない
     * かつ Listクラスにremoveメソッドを多用することで配列コピーを多用されることを避けるため
	 * （重くなることをできるだけ避ける）
	 * </pre>
	 * @param layout スプライト層
	 * @param sprite スプライト
	 */
	public void removeMediaView(int layout, KSoundMP3 mp3){

		// スプライト一覧を取得
		List<Node> spriteLst = this.getSpriteLst(layout);

		// インデックスを求める
		int index = spriteLst.indexOf(mp3);

		// インデックスが取得できた場合、
		if(index != -1){

			// ダミースプライトに置き換え
			spriteLst.set(index, new KSpriteDummy(this));

		}

	} // end removeMediaView

	/**
	 * キャンバス追加
	 * @param layout スプライト層
	 * @param kCanvas キャンバスクラス
	 */
	public void addCanvas(int layout, KCanvas kCanvas){

		// スプライト一覧を取得
		List<Node> spriteLst = this.getSpriteLst(layout);

		// キャンバスクラスを追加
		spriteLst.add(kCanvas);

	} // end addCanvas

	/**
	 * キャンバス削除
	 * <pre>
	 * スプライト層から探索して、キャンバスを削除します。
	 * 当メソッドでは、ダミースプライトに置き換え、runメソッドの中で削除します。
	 * </pre>
	 * @param layout スプライト層
	 * @param kCanvas キャンバスクラス
	 */
	public void removeCanvas(int layout, KCanvas kCanvas){

		// スプライト一覧を取得
		List<Node> spriteLst = this.getSpriteLst(layout);

		// インデックスを取得
		int index = spriteLst.indexOf(kCanvas);

		// インデックスが取得できた場合、
		if(index != -1){

			// ダミースプライトに置き換え
			spriteLst.set(index, new KSpriteDummy(this));

		}

	} // end removeCanvas

	/**
	 * キャンバス削除
	 * <pre>
	 * 全てのスプライト層を探索して、キャンバスを削除します。
	 * 通常はスプライト層が引数として設定しているメソッドを呼んでください。
	 * </pre>
	 * @param canvas キャンバス
	 */
	public void removeCanvas(KCanvas canvas){

		// スプライト層深さ分、ループ
		for(int i = 0; i < spriteLayoutCount; i++){

			// スプライト削除
			removeCanvas(i, canvas);

		}

	} // end addCanvas

	/**
	 * コントロールクラス取得
	 * @return KCtrl コントロールクラス
	 */
	public KCtrl getKCtrl(){

		// コントロールクラスをリターン
		return ctrl;

	} // end getKCtrl

	/**
	 * 画面伸縮モード取得
	 * <pre>
	 * 画面の伸縮に合わせて、画像も伸縮させるようにするか／しないかを返します。
	 * </pre>
	 * @return boolean true：伸縮させる false：伸縮させない
	 */
	public boolean getStretchScreen(){
		return this.stretchScreen;
	}

	/**
	 * 画面伸縮モード設定
	 * <pre>
	 * 画面の伸縮に合わせて、画像も伸縮させるようにするか、決定します。
	 * </pre>
	 * @param stretchScreen true：伸縮させる false：伸縮させない
	 */
	public void setStretchScreen(boolean stretchScreen){
		this.stretchScreen = stretchScreen;
	}

	/**
	 * 比率 取得
	 * @return double 比率
	 */
	public double getRatio(){
		return this.ratio;
	}

	/**
	 * 比率 設定
	 * @param ratio 比率
	 */
	public void getRatio(double ratio){
		this.ratio = ratio;
	}

	/**
	 * キー押下中かをどうかをチェック
	 * @param code キーコード
	 * @return boolean true：押している ／ false：押していない
	 */
	public boolean isKeyPressed(KeyCode code){

		return mapKeyPress.containsKey(code);

	} // end isKeyPressed

	/**
	 * キーを押したイベント
	 * <pre>
	 * ラムダ式記述を行って、インターフェース作成を省略しています。
	 * </pre>
	 */
	public void KPanelKeyPressed(Event e){

		KeyEvent ke = (KeyEvent)e;
		KeyCode code = ke.getCode();
		mapKeyPress.put(code, "");

	} // end KPanelKeyPressed

	/**
	 * キーを離したイベント
	 * <pre>
	 * ラムダ式記述を行って、インターフェース作成を省略しています。
	 * </pre>
	 */
	public void KPanelKeyKeyReleased(Event e){

		KeyEvent ke = (KeyEvent)e;
		KeyCode code = ke.getCode();
		mapKeyPress.remove(code);

	} // end KPanelKeyKeyReleased

}
